from __future__ import absolute_import
from __future__ import unicode_literals

import errno
import socket
from socket import getfqdn
from socket import SO_REUSEADDR
from socket import socket as Socket
from socket import SOL_SOCKET

from ephemeral_port_reserve import LOCALHOST
from ephemeral_port_reserve import reserve


def bind_naive(ip, port):
    sock = Socket()
    try:
        sock.bind((ip, port))
    except socket.error as error:
        return error


def bind_reuse(ip, port):
    sock = Socket()
    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    sock.bind((ip, port))
    return sock


def assert_ip(ip):
    port = reserve(ip)

    # show that we can't bind to it without SO_REUSEADDR
    error = bind_naive(ip, port)
    assert error and error.errno == errno.EADDRINUSE, error

    # show that we *can* bind to it without SO_REUSEADDR, after release
    sock = bind_reuse(ip, port)
    sname = sock.getsockname()
    assert sname == (ip, port), (sname, port)


def test_localhost():
    assert_ip(LOCALHOST)


def test_fqdn():
    fqip = socket.gethostbyname(getfqdn())
    assert_ip(fqip)


def test_preferred_port():
    import ephemeral_port_reserve
    print(ephemeral_port_reserve)
    port = reserve()
    port2 = reserve(port=port)
    assert port == port2
    assert bind_reuse(LOCALHOST, port2)


def test_preferred_port_in_use():
    """if preferred port is in use, it will find an unused port"""
    port = reserve()
    sock = bind_reuse(LOCALHOST, port)
    sock.listen(1)  # make the port in-use
    port2 = reserve(port=port)
    assert port != port2
    assert bind_reuse(LOCALHOST, port2)



#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import unicode_literals

import contextlib
import errno
from socket import error as SocketError
from socket import SO_REUSEADDR
from socket import socket
from socket import SOL_SOCKET

LOCALHOST = '127.0.0.1'


def reserve(ip=LOCALHOST, port=0):
    """Bind to an ephemeral port, force it into the TIME_WAIT state, and unbind it.

    This means that further ephemeral port alloctions won't pick this "reserved" port,
    but subprocesses can still bind to it explicitly, given that they use SO_REUSEADDR.
    By default on linux you have a grace period of 60 seconds to reuse this port.
    To check your own particular value:
    $ cat /proc/sys/net/ipv4/tcp_fin_timeout
    60

    By default, the port will be reserved for localhost (aka 127.0.0.1).
    To reserve a port for a different ip, provide the ip as the first argument.
    Note that IP 0.0.0.0 is interpreted as localhost.
    """
    port = int(port)
    with contextlib.closing(socket()) as s:
        s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        try:
            s.bind((ip, port))
        except SocketError as e:
            # socket.error: EADDRINUSE Address already in use
            if e.errno == errno.EADDRINUSE and port != 0:
                s.bind((ip, 0))
            else:
                raise

        # the connect below deadlocks on kernel >= 4.4.0 unless this arg is greater than zero
        s.listen(1)

        sockname = s.getsockname()

        # these three are necessary just to get the port into a TIME_WAIT state
        with contextlib.closing(socket()) as s2:
            s2.connect(sockname)
            sock, _ = s.accept()
            with contextlib.closing(sock):
                return sockname[1]


def main():  # pragma: no cover
    from sys import argv
    port = reserve(*argv[1:])
    print(port)


if __name__ == '__main__':
    exit(main())
